#include "BitmapImage.h"

BitmapImage::BitmapImage( int w, int h ) {

   width = w;
   height = h;

   redChannel = new Matrix<unsigned char>( w, h, 0xff );
   greenChannel = new Matrix<unsigned char>( w, h, 0xff );
   blueChannel = new Matrix<unsigned char>( w, h, 0xff );

};

BitmapImage::~BitmapImage() {

   if (redChannel != 0) delete redChannel;
   if (greenChannel != 0) delete greenChannel;
   if (blueChannel != 0) delete blueChannel;
}

unsigned short BitmapImage::readShort( ifstream &input ) {

   unsigned short result;
   char byte[2];

   input.read( byte, 2 );

   result = *((unsigned short *)&byte);

   return result;
}

unsigned long BitmapImage::readLong( ifstream &input ) {

   unsigned long result;
   char byte[4];

   input.read( byte, 4 );

   result = *((unsigned long *)&byte);

   return result;
}

void BitmapImage::writeShort( unsigned short s, ofstream &output ) {

   const char *byte = ((char *)&s);

#ifdef WIN32
   output.write( byte, 2 );
#else
   
   output.write( byte+1, 1 );
   output.write( byte, 1 );
   
#endif
}

void BitmapImage::writeLong( unsigned long l, ofstream &output ) {

   const char * byte = (char *)&l;

#ifdef WIN32
   output.write( byte, 4 );
#else
   
   output.write( byte+3, 1 );
   output.write( byte+2, 1 );
   output.write( byte+1, 1 );
   output.write( byte, 1 );
   
#endif
}

BitmapImage::BitmapImage( ifstream &input ) {

   // Read in the identifier.
   input.read( identifier, 2 );

   // Read the filesize.
   fileSize = readLong( input );
   reserved = readLong( input );
   offset = readLong( input );
   headerLength = readLong( input );
   width = readLong( input );
   height = readLong( input );
   planes = readShort( input );
   bitsPerPixel = readShort( input );
   compression = readLong( input );
   bitmapSize = readLong( input );
   hResolution = readLong( input );
   vResolution = readLong( input );
   colors = readLong( input );
   importantColors = readLong( input );

   // Make sure image is 24 bit.
   if (bitsPerPixel != 24) {
      assert( false && "Not a 24 Bit image" );
   }

   // Initalize each of three channels.
   redChannel = new Matrix< unsigned char >( height, width );
   blueChannel = new Matrix< unsigned char >( height, width );
   greenChannel = new Matrix< unsigned char >( height, width );

   // Read in data.
   int bytesPerRow = ((3 * width + 3)/4) * 4;
   int padding = bytesPerRow - 3 * width;

   char byte[3];

   for (unsigned int i=0;i<height;++i) {
      for (unsigned int j=0;j<width;++j) {
            
         input.read( byte, 3 );

         blueChannel->get( i, j) = byte[0];
         greenChannel->get( i, j) = byte[1];
         redChannel->get( i, j) = byte[2];
      }

      // Move over padding.
      for (int k=0;k<padding;++k) input.read( byte, 1);
   }
}

void BitmapImage::output( ofstream &out ) {

   offset = 54;
   fileSize = offset + bitmapSize;
   headerLength = 40;
   width = redChannel->getRows();
   height = redChannel->getColumns();
   planes = 1;
   bitsPerPixel = 24;
   compression = 0;
   bitmapSize = width * height * 3;
   hResolution = vResolution = colors = importantColors = 0;

   out.write( "BM", 2 );
   
   writeLong( fileSize, out );

   // Reserved.
   writeLong( 0, out );   
   
   writeLong( offset, out );
   writeLong( headerLength, out );
   writeLong( width, out );
   writeLong( height, out );
   writeShort( planes, out );
   writeShort( bitsPerPixel, out );
   writeLong( compression, out );
   writeLong( bitmapSize, out );
   writeLong( hResolution, out );
   writeLong( vResolution, out );
   writeLong( colors, out );
   writeLong( importantColors, out );

   // Write the data.
   int bytesPerRow = ((3 * width + 3)/4) * 4;
   int padding = bytesPerRow - 3 * width;
   char zero = 0;

   for (unsigned int i=0;i<width;++i) {
      for (unsigned int j=0;j<height;++j) {

         char bgr[3];

         bgr[0] = blueChannel->get(i,j);
         bgr[1] = greenChannel->get(i,j);
         bgr[2] = redChannel->get(i,j);

         out.write( bgr, 3 );
      }

      for (int k=0;k<padding;++k) out.write( &zero, 1 );
   }

}

inline void BitmapImage::set( int x, int y, Color pixel ) {

   // Re-orient so that 0,0 is in the upper left, it is in the 
   // lower left in Bitmaps by default.
   x = (getRows() - 1) - x;
   
   // Make sure that no wrap around occurs.
   pixel.bound( 0.0, 1.0 );

   // Set the image to the color of the pixel.
   redChannel->get( x, y ) = (unsigned char)((255) * pixel.r);
   greenChannel->get( x, y ) = (unsigned char)((255) * pixel.g);
   blueChannel->get( x, y ) = (unsigned char)((255) * pixel.b);
}

void BitmapImage::set( int i, int j, unsigned char r, unsigned char g, unsigned char b ) {

   redChannel->get( i, j ) = r;
   greenChannel->get( i, j ) = g;
   blueChannel->get( i, j ) = b;
}





