#ifndef MATRIX__H
#define MATRIX__H

#include <assert.h>
#include <stdio.h>
#include <iostream>

using namespace std;

#include "Vector.h"


template< class T >
class Matrix {
public:
   int rows;
   int columns;
   T * data;

public:
   Matrix();
   Matrix( int r, int c );
   Matrix( int r, int c, T initial );
   Matrix( const Matrix< T > &matrix );

   ~Matrix();

   // Accessors.
   inline T &get( int i, int j );
   inline T constGet( int i, int j ) const;
   T &safeGet( int i, int j);
   inline T &get( int i );
   inline const int getRows() { return rows; };
   inline const int getColumns() { return columns; };
   Matrix< T > &operator =( const Matrix< T > &matrix );

   // Matrix operations.
   Matrix &operator *= ( T scalar );
   Matrix &operator += ( const Matrix &matrix );
   Matrix &operator += ( T scalar );

   // Binary Operators.
   Matrix< T > operator * ( const Matrix< T > &b ) const;
   Matrix< T > operator * ( T scalar ) const;
   Vector operator * (const Vector v) const;
   Matrix< T > operator + ( const Matrix< T > &matrix ) const;
   Matrix< T > operator + ( T scalar ) const;
   Matrix< T > operator - ( const Matrix< T > &matrix ) const;
   Matrix< T > operator - ( T scalar ) const;

   // Other operations.
   Matrix transpose() const;
   Matrix &convolve( Matrix< T > &kernel );
   T sumElements();
   Matrix scale( const Matrix< T > &matrix );

   void outputFormatted( ostream &out );
   Matrix< float > floatMatrix();
   Matrix< unsigned char > ucharMatrix();
};

template< class T >
Matrix< T >::Matrix() {
   rows = columns = 0;
   data = 0;
}

template< class T >
Matrix< T >::Matrix( int r, int c ) : rows( r ), columns( c ) {

   data = new T[ rows * columns ];
}

template< class T >
Matrix< T >::Matrix( int r, int c, T initial ) : rows( r ), columns( c ) {

   data = new T[ rows * columns ];

   for (int i=0;i<rows*columns;++i) {
      data[i] = initial;
   }
}
/*
template< class T >
Matrix< T >::Matrix( const Matrix< unsigned char > &matrix ) {

   rows = matrix.rows;
   columns = matrix.columns;

   data = new T[ rows * columns ];

   for (int i=0;i<rows*columns; ++i) {
      data[i] = matrix.data[i];
   }
}
*/
template< class T >
Matrix< T >::Matrix( const Matrix< T > &matrix ) {

   rows = matrix.rows;
   columns = matrix.columns;

   data = new T[ rows * columns ];

   for (int i=0;i<rows*columns; ++i) {
      data[i] = matrix.data[i];
   }
}

template< class T >
Matrix< T >::~Matrix() {

   if (data != 0) {
      delete data;
   }
}

// Accessors
template< class T >
inline T &Matrix< T >::get( int i, int j ) {

   assert( i < rows && j < columns && i >= 0 && j >= 0 );

   return data[((i*columns)+j)];
}

template< class T >
inline T Matrix< T >::constGet( int i, int j ) const {

   assert( i < rows && j < columns && i >= 0 && j >= 0 );

   return data[((i*columns)+j)];
}

template< class T >
T &Matrix< T >::safeGet( int i, int j ) {

   if (i<0) i = 0;
   if (i>=rows) i = rows-1;
   if (j<0) j = 0;
   if (j>=columns) j = columns-1;

   return data[((i*columns)+j)];
}


template< class T >
inline T &Matrix< T >::get( int i ) {

   assert( i < rows * columns && i >= 0 );

   return data[i];
}

// Assignment operators.
/*
Matrix< float > &Matrix< float >::operator =( const Matrix< unsigned char > &matrix ) {
   
   if (data != 0) {
      delete data;
   }

   rows = matrix.getRows();
   columns = matrix.getColumns();

   data = new float[ rows * columns ];

   for (int i=0;i<rows*columns; ++i) {
      data[i] = matrix.get(i);
   }

   return *this;
}
*/

template< class T >
Matrix< T > &Matrix< T >::operator =( const Matrix< T > &matrix ) {
   
   if (data != 0) {
      delete data;
   }

   rows = matrix.rows;
   columns = matrix.columns;

   data = new T[ rows * columns ];

   for (int i=0;i<rows*columns; ++i) {
      data[i] = matrix.data[i];
   }

   return *this;
}

// Basic operations.
template< class T >
Matrix< T > &Matrix< T >::operator *= ( T scalar ) {

   for (int i=0;i<rows*columns;++i) {
      data[i] *= scalar;
   }

   return *this;
}

template< class T >
Matrix< T > &Matrix< T >::operator += ( const Matrix< T > &matrix ) {

   assert( (rows == matrix.rows) && (columns == matrix.columns) );

   for (int i=0;i<rows*columns;++i) {
      data[i] += matrix.data[i];
   }

   return *this;
}

template< class T >
Matrix< T > &Matrix< T >::operator += ( T scalar ) {

   for (int i=0;i<rows*columns;++i) {
      data[i] += scalar;
   }

   return *this;
}


template< class T >
Matrix< T > Matrix<T>::operator * ( const Matrix< T > &b ) const {

   Matrix< T > r( rows, b.columns, 0.0 );
   int i, j,s;

   assert( rows == b.columns );

   for (j=0;j<columns;++j) {
      for (i=0;i<b.rows;++i) {
         for (s=0;s<rows;++s) {
            r.get( i, j ) += (constGet( i, s ) * b.constGet( s, j ));
         }
      }
   }

   return r;
}

template< class T>
Matrix< T > Matrix<T>::operator * ( T scalar ) const {

   Matrix< T > r( rows, columns );

   for (int i=0;i<rows*columns;++i) {
      r.data[i] = data[i] * scalar;
   }

   return r;
}

template< class T >
Vector Matrix<T>::operator * (const Vector v) const {

   assert( rows == 4 && columns == 4 );

   Vector result;
   float h;

   result.x = data[0] * v.x +
              data[1] * v.y +
              data[2] * v.z +
              data[3];

   result.y = data[4] * v.x +
              data[5] * v.y +
              data[6] * v.z +
              data[7];

   result.z = data[8] * v.x +
              data[9] * v.y +
              data[10] * v.z +
              data[11];

   h = data[12] * v.x +
       data[13] * v.y +
       data[14] * v.z +
       data[15];

   result *= (1.0F/h);

   return result;
}

template< class T >
Matrix< T > Matrix<T>::operator + ( const Matrix< T > &matrix ) const {

   assert( rows == matrix.rows && columns == matrix.columns );

   Matrix< T > r( rows, columns );

   for (int i=0;i<rows*columns;++i) {
      r.data[i] = data[i] + matrix.data[i];
   } 

   return r;
}

template< class T >
Matrix< T > Matrix<T>::operator + ( T scalar ) const {

   Matrix< T > r( rows, columns );
   
   for (int i=0;i<rows*columns;++i) {
      r.data[i] = data[i] + scalar;
   }

   return r;
}

template< class T >
Matrix< T > Matrix<T>::operator - ( const Matrix< T > &matrix ) const {

   assert( rows == matrix.rows && columns == matrix.columns );

   Matrix< T > r( rows, columns );

   for (int i=0;i<rows*columns;++i) {
      r.data[i] = data[i] - matrix.data[i];
   } 

   return r;
}

template< class T >
Matrix< T > Matrix<T>::operator - ( T scalar ) const {

   assert( rows == matrix.rows && columns == matrix.columns );

   Matrix< T > r( rows, columns );
   
   for (int i=0;i<rows*columns;++i) {
      r.data[i] = data[i] - scalar;
   }

   return r;
}

template< class T >
Matrix< T > &Matrix< T >::convolve( Matrix< T > &kernel ) {

   int radius = kernel.rows / 2;
   Matrix< T > r( rows, columns );

   for (int i=0;i<rows;++i) {
      for (int j=0;j<columns;++j) {

         T total = 0;

         // Box around the ith element of this.
         for (int k=-1*radius;k<radius;++k) {
            for (int l=-1*radius;l<radius;++l) {
               
               T weight = kernel.get( k+radius, l+radius );

               total += weight * safeGet( i+k, j+k  );
            }
         }

         r.get( i, j ) = total;
      }
   }

   *this = r;

   return *this;
};

template< class T >
void Matrix< T >::outputFormatted( ostream &out ) {

   for (int j=0;j<columns;++j) {
      for (int i=0;i<rows;++i) {
         // out << "0x" << hex << data[(rows*j)+i] << dec << " ";
         out << data[(rows*j)+i] << " ";
      }
      out << endl;
   }


}

template< class T >
T Matrix< T >::sumElements() {

   T sum = 0;
   for (int i=0;i<rows*columns;++i) {
      sum += data[i];
   }

   return sum;
}

template< class T >
Matrix< T > Matrix< T >::scale( const Matrix< T > &matrix ) {

   assert( rows == matrix.rows && columns == matrix.columns );

   Matrix< T > r( rows,columns );

   for (int i=0;i<rows*columns; ++i) {
      r.data[i] = data[i] * matrix.data[i];
   }

   return r;
}

template< class T >
Matrix< float > Matrix< T >::floatMatrix() {

   Matrix< float > r( rows, columns );

   for (int i=0;i<rows*columns;++i) {
      r.get(i) = data[i];
   }

   return r;
}

template< class T >
Matrix< unsigned char > Matrix< T >::ucharMatrix() {

   Matrix< unsigned char > r( rows, columns );

   for (int i=0;i<rows*columns;++i) {
      r.get(i) = data[i];
   }

   return r;
}

template< class T >
Matrix< T > Matrix< T >::transpose() const {

   Matrix< T > t( columns, rows );

   for (int j=0;j<columns;++j) {
      for (int i=0;i<rows;++i) {
         t.get( i, j ) = constGet( j, i );
      }
   }

   return t;
}

#endif
