#ifndef VECTOR__H
#define VECTOR__H

#include "math.h"

class Vector {
public:
   float x;
   float y;
   float z;

   Vector() { x = y = z = 0.0; };
   Vector( float _x, float _y, float _z ) { x = _x; y = _y; z = _z; };
   Vector( const Vector &v ) { x = v.x; y = v.y; z = v.z; }

   // Addition.
   inline Vector &operator += ( const Vector &v ) { x += v.x; y += v.y; z += v.z; return *this; };
   inline Vector &operator += ( float f ) { x += f; y += f; z += f; return *this; };
   inline Vector operator + ( float f ) const { return Vector( x + f, y + f, z + f ); };
   inline Vector operator + ( const Vector &v ) const { return Vector( x + v.x, y + v.y, z + v.z ); }
   inline Vector operator - ( const Vector &v ) const { return Vector( x - v.x, y - v.y, z - v.z );  };

   // Multiplication and scaling.
   inline Vector &operator *= ( const Vector &v ) { x *= v.x; y *= v.y; z *= v.z; return *this; };
   inline Vector &operator *= ( float f ) { x *= f; y *= f; z *= f; return *this;};
   inline Vector operator * ( const Vector &v ) const { return Vector( x * v.x, y * v.y, z * v.z ); };
   inline Vector operator * ( float f ) const { return Vector( x * f, y * f, z * f ); }

   // Equality, assignment.
   inline bool operator == ( const Vector &v ) const { return x == v.x && y == v.y && z == v.z; };
   inline bool operator != ( const Vector &v ) const { return !(*this == v ); }
   inline Vector &operator = ( const Vector &v ) { x = v.x; y = v.y; z = v.z; return *this; };

   // Vector operations.
   inline Vector &normalize() { float len = (float)sqrt( x*x + y*y + z*z ); if (len != 1.0) { x /= len; y /= len; z /= len; } return *this; }
   inline Vector constNormalize() const { Vector v = *this; return v.normalize(); };
   inline float length() const { return (float)sqrt( x*x + y*y + z*z ); };
   inline float dot( const Vector &v ) const { return x*v.x + y*v.y + z*v.z; };
   inline Vector cross( const Vector &v ) const { return Vector( y * v.z - z * v.y, z * v.x - x * v.z, x * v.y - y * v.x ); };
   inline Vector abs() const { return Vector( (float)fabs(x), (float)fabs(y), (float)fabs(z) ); }

   inline float &operator [] (const unsigned int index) { if (index==0) return x; else if (index==1) return y; return z; };
   inline float operator [] (const unsigned int index) const { if (index==0) return x; else if (index==1) return y; return z; };
};

class Color {
public:
   float r;
   float g;
   float b;
   
   Color() { r = g = b = 0.0F; };
   Color( float _r, float _g, float _b ) { r = _r; g = _g; b = _b; };
   Color( const Color &c ) { r = c.r; g = c.g; b = c.b; };
   
   // Color addition.
   inline Color &operator += ( const Color &c ) { r += c.r; g += c.g; b += c.b; return *this; }
   inline Color operator + ( const Color &c ) { return Color( r + c.r, g+c.g, b+c.b ); };

   // Equality, assignment.
   inline bool operator == ( const Color &v ) { return r == v.r && g == v.g && b == v.b; };
   inline Color &operator = ( const Color &v ) { r = v.r; g = v.g; b = v.b; return *this; };

   // Scaling and elementwise multiplication.
   inline Color &operator *= ( const Color &c ) { r *= c.r; g *= c.g; b *= c.b; return *this; }
   inline Color &operator *= ( float f ) { r *= f; g *= f; b *= f; return *this; };
   inline Color operator * ( const Color &c ) const { return Color( r * c.r, g * c.g, b * c.b ); };
   inline Color operator * ( float f ) { return Color( r * f, g * f, b * f ); };

   inline Color &bound( float min, float max ){ if (r > max) r = max; if (g > max) g = max; if (b > max) b = max;
                                                if (r < min) r = min; if (g < min) g = min; if (b < min) b = min;
                                                return *this; }
};

class Ray {
public:
   Vector position;
   Vector direction;
   int i;
   int j;

   Ray( const Vector &p, const Vector &d, int _i, int _j ) : 
      position( p ), direction( d ), i(_i), j(_j) { direction.normalize(); };
   Ray( const Vector &p, const Vector &d ) : 
      position( p ), direction( d ), i(-1), j(-1) { direction.normalize(); };
};

#endif